//-- I N C L U D E S ------------------------
#include <p18cxxx.h>
#include "system\typedefs.h"
#include "system\usb\usb.h"
#include "io_cfg.h"
#include "user\user.h"
#include <timers.h>
#include <string.h>
#include <delays.h>

//-- V A R I A B L E S ----------------------

#pragma udata
byte colors[3] = {0,0,0};
byte real_colors[3] = {0,0,0};
byte white_balance[3] = {245,255,195};

#define OSC_BYTES 63
#define OSC_VALUES 48
#define OSC_ARRAYS 18
//#define OSC_VALUES 25

byte osc_fill, osc_sent, osc_pos, osc=0, osc_overflow, osc_seq, osc_packets;

word osc_sleeptime = 0;

//pun  =  0.36  0.40  2x3R9 -> 0.39
//vihr =  0.35  0.31  (1R tai 1R2 -> 0.37) 2x0.68 -> 0.35
//sin  =  0.19  3R9 > 0.24  6R6 > 0.19 R68 > 0.33

//(vihr: 0 > 0.48A, sin 0 > 0.41A, pun: 3R3+1R0 > 0.5A, tot 0,97A

// sin  = 0.29  0.35
// vihr = 0.32  0.35
// pun  = 0.41  0.40

#pragma udata
#define rpin LATBbits.LATB5
#define gpin LATBbits.LATB4
#define bpin LATBbits.LATB3

#define dpin LATBbits.LATB7
#define uint word
byte counter;
#pragma udata gpr2
DATA_PACKET dataPacket;
#pragma udata

void ServiceRequests(void);


#pragma udata gpr0
static char oscillator_1[OSC_BYTES]; // oscilloscope arrays
static char oscillator_2[OSC_BYTES]; // oscilloscope arrays

#pragma udata gpr1
static char oscillator_3[OSC_BYTES]; // oscilloscope arrays
static char oscillator_4[OSC_BYTES]; // oscilloscope arrays
static char oscillator_5[OSC_BYTES]; // oscilloscope arrays
static char colorArray[OSC_BYTES]; // two oscilloscope arrays

#pragma udata usbram5=0x500
static char oscillator_6[OSC_BYTES]; // oscilloscope arrays
static char oscillator_7[OSC_BYTES]; // oscilloscope arrays
static char oscillator_8[OSC_BYTES]; // oscilloscope arrays
static char oscillator_9[OSC_BYTES]; // oscilloscope arrays

#pragma udata usbram6=0x600
static char oscillator_10[OSC_BYTES]; // oscilloscope arrays
static char oscillator_11[OSC_BYTES]; // oscilloscope arrays
static char oscillator_12[OSC_BYTES]; // oscilloscope arrays
static char oscillator_13[OSC_BYTES]; // oscilloscope arrays

#pragma udata usbram7=0x700
static char oscillator_14[OSC_BYTES]; // oscilloscope arrays
static char oscillator_15[OSC_BYTES]; // oscilloscope arrays
static char oscillator_16[OSC_BYTES]; // oscilloscope arrays
static char oscillator_17[OSC_BYTES]; // oscilloscope arrays


#pragma udata

#pragma code

char * cols = &colorArray[0];


void InterruptHandlerHigh (void);
void InterruptHandlerLow (void);

static int state = 0;
static uint scaler = 0;
static uint endpoint_wait_counter = 0;

uint colam = 3;
uint speed = 10;
uint endpoint_wait_time = 10;

void waitx(int time) {
  int i;
  for(i=0; i< time; i++) {
//    for(k=0; k< time2; k++) {
      Delay1TCY();    
//    }
  }    
}

#pragma code InterruptVectorHigh = 0x08
void InterruptVectorHigh() {
  _asm
    goto InterruptHandlerHigh //jump to interrupt routine
  _endasm
}

#pragma code InterruptVectorLow = 0x18
void InterruptVectorLow() {
  _asm
    goto InterruptHandlerLow //jump to interrupt routine
  _endasm
}

#pragma code


char * getOscArray(byte num) {
// I didn't find a better way to do this as... 
// char * oscArrays = { &oscillator_2[0], &oscillator_3[0], &oscillator_4[0], &colorArray[0]...
// ..didn't work...
    switch (num) {
        case 0:
            return &oscillator_1[0];
        case 1:
            return &oscillator_2[0];
        case 2:
            return &oscillator_3[0];
        case 3:
            return &oscillator_4[0];
        case 4:
            return &colorArray[0];
        case 5:
            return &oscillator_5[0];
        case 6:
            return &oscillator_6[0];
        case 7:
            return &oscillator_7[0];
        case 8:
            return &oscillator_8[0];
        case 9:
            return &oscillator_9[0];
        case 10:
            return &oscillator_10[0];
        case 11:
            return &oscillator_11[0];
        case 12:
            return &oscillator_12[0];
        case 13:
            return &oscillator_13[0];
        case 14:
            return &oscillator_14[0];
        case 15:
            return &oscillator_15[0];
        case 16:
            return &oscillator_16[0];
        case 17:
            return &oscillator_17[0];
    }    
}




#define getRealColor(col) ((byte)( ((uint)colors[col]*(uint)white_balance[col]) >> 8)) 

void advanceColors() {
    char i;
    scaler++;
    if (scaler > (uint)speed*speed) {
      scaler = 0;

      if (endpoint_wait_time*endpoint_wait_time < endpoint_wait_counter) {
        endpoint_wait_counter = 0;
      }
      if (endpoint_wait_counter > 0) {
        endpoint_wait_counter++;
      } 
      else {
   
        for(i=0; i< 3; i++) {
          if (cols[state*3+i] > colors[i]) {
            colors[i]++;
          }
          if (cols[state*3+i] < colors[i]) {
            colors[i]--;
          }
        }
        //setWBal();
        for(i=0; i< 3; i++) {
          real_colors[i] = getRealColor(i);//realCols[i][colors[i]];
        }
        //coladv++;
        if ( cols[state*3] == colors[0] && cols[state*3+1] == colors[1]  && cols[state*3+2] == colors[2] ) {
          state++;
          if (state >= colam) { 
            state = 0;
          }
          endpoint_wait_counter = 1;
        }
        
      }

    }
}


char * current_osc;
#pragma interruptlow InterruptHandlerLow
void InterruptHandlerLow(){
    byte bytepos, bitpos,i;

  if (PIR1bits.TMR2IF) {
      PIR1bits.TMR2IF = 0;
          advanceColors();
      
  }
  


  if (PIR1bits.ADIF) {
      PIR1bits.ADIF = 0;
      if (osc) {
          //save value
          if (osc_pos >= OSC_VALUES) {
              osc_pos = 0;
              if (osc_sent == osc_fill) {
                  osc_overflow = 1;
                  current_osc[OSC_BYTES-1] = 1;
              }
              osc_fill++;
              if (osc_fill >= OSC_ARRAYS)
                  osc_fill = 0;
              osc_seq++;
              current_osc = getOscArray(osc_fill);


          }
          
          // sent   fill  
          //   7      0     
          //   7      0     
          //   7      1     
          //   7      1     
          //   0      1     
          //   0      1     
          //   0      2     
          //   0      2     
          //   1      2     
          //   1      2     
          //   1      3     
          //   1      3     
          //   2      3     
          //   2      3     
          
          if (!osc_overflow) {
              if (osc_pos == 0) {
                  for(i=0; i<OSC_BYTES;i++) {
                      current_osc[i] = 0;
                  }
              }
              current_osc[osc_pos] = ADRESL;
              bytepos = OSC_VALUES + (osc_pos >> 2);
              bitpos  = (osc_pos%4)*2;
              current_osc[bytepos] = current_osc[bytepos] | (ADRESH << bitpos);
              osc_pos++;
              ADCON0bits.GO = 1;
          }

      }
  }
  
}

byte getBufferState() {
    // osc_fill  osc_sent  bufferState
    //    1         0          1
    //    3         0          3
    //    3         1          2
    //    0         BA-1(7)    1
    //    0         BA-2(6)    2
    //    3         BA-2(6)    5
    
    
    if (osc_fill > osc_sent)
      return osc_fill - osc_sent;
      
    // osc_fill <= osc_sent
    return OSC_ARRAYS + osc_fill - osc_sent;
    
}

#pragma code

char * sending_osc;
void oscilloscope(void) {
    byte sendArray;
    if (osc) {
        if ( getBufferState() > 1) {

            sendArray = osc_sent+1;
            if (sendArray >= OSC_ARRAYS)
                sendArray = 0;

            sending_osc = getOscArray(sendArray); //oscArrays[1-osc_sent];

                
            while(mUSBGenTxIsBusy()) { }
            USBGenWrite((byte*)sending_osc, OSC_BYTES);
            if (osc_overflow) {
                osc_overflow = 0;
                ADCON0bits.GO = 1;
            }
            osc_sent = sendArray;

        }
    }
}





void setpin(byte pin, byte state){
    switch(pin) {
        case 0:
            rpin = state;
            break;
        case 1:
            gpin = state;
            break;
        case 2:
            bpin = state;
            break;
    }    
}

byte getpin(byte pin){
    switch(pin) {
        case 0:
            return rpin;
            break;
        case 1:
            return gpin;
            break;
        case 2:
            return bpin;
            break;
    }    
}


void writeTimer(byte timer, int time) {
    switch(timer) {
        case 0:
            WriteTimer0(time);
            break;
        case 1:
            WriteTimer1(time);
            break;
        case 3:
            WriteTimer3(time);
            break;
    }    
}



uint min(uint a, uint b) {
    if (a < b)
        return a;
    return b;
}

void colorInterrupt(byte pin,byte timer, byte color) {
    if (getpin(pin)) {
        setpin(pin,0);
        writeTimer(timer, (uint)color*color);
    } else {
        setpin(pin, 1);
        writeTimer(timer, 65426-(uint)color*color );
    }



    
}




#pragma interrupt InterruptHandlerHigh
void InterruptHandlerHigh(){
    
  if (INTCONbits.TMR0IF) {              //check for TMR overflow
    INTCONbits.TMR0IF = 0;            //clear interrupt flag
    colorInterrupt(1, 0, real_colors[1]);
  }

  if (PIR1bits.TMR1IF) {            //check for TMR overflow
    PIR1bits.TMR1IF = 0;            //clear interrupt flag
    colorInterrupt(0, 1, real_colors[0]);
  }

  if (PIR2bits.TMR3IF) {            //check for TMR overflow
    PIR2bits.TMR3IF = 0;            //clear interrupt flag
    colorInterrupt(2, 3, real_colors[2]);
  }
  


}


#pragma code
byte read_EEPROM(byte address) {
    
    EEADR = address;
    EECON1bits.EEPGD = 0;
    EECON1bits.CFGS = 0;
    EECON1bits.RD = 1;
    return EEDATA;
}


void write_EEPROM(byte address, byte data) {
    if (read_EEPROM(address) == data) {
        return;
    }

    EECON1bits.EEPGD = 0;
    EECON1bits.CFGS = 0;
    EECON1bits.WREN = 1;
    EEADR = address;
    EEDATA = data;
    INTCONbits.GIE = 0;
    EECON2 = 0x55;
    EECON2 = 0xaa;
    EECON1bits.WR = 1;
    INTCONbits.GIE = 1;
    while (EECON1bits.WR) {}
    EECON1bits.WREN = 0;
}


void updateEEPROM() {
    int i;
    write_EEPROM(0, speed);
    write_EEPROM(1, endpoint_wait_time);
    write_EEPROM(2, colam);
    for(i=0; i< colam*3; i++) {
        write_EEPROM(i+3, cols[i]);   
    }
}


void ot2(unsigned char config) {
  T2CON = (0xfb & config);  // Set all configuration values, but
                            // don't start timer yet

  TMR2 = 0;                 // Clear Timer2
  PIR1bits.TMR2IF = 0;

  if(config & 0x80)         // Enable timer interrupts?
    PIE1bits.TMR2IE = 1;
  else
    PIE1bits.TMR2IE = 0;

  T2CONbits.TMR2ON = 1; // Turn on Timer2
}


void UserInit(void)
{
    int i;
    TRISB = 0b00000011;
    dpin = 1;
    
    //Read stuff from EEPROM
    speed = read_EEPROM(0);
    endpoint_wait_time = read_EEPROM(1);
    colam = read_EEPROM(2);
    if (colam > MAX_COL_AM) {
        speed = 0;
        endpoint_wait_time = 0;
        colam = 2;
        cols[0] = 0;
        cols[1] = 0;
        cols[2] = 0;
        cols[3] = 255;
        cols[4] = 255;
        cols[5] = 255;
    }
    else {
        for(i=0; i< colam*3; i++) {
            cols[i] = read_EEPROM(i+3);
        }
    }
    
    
    if (read_EEPROM(252) == 178) {
      for(i=0; i< 3; i++) {
        white_balance[i] = read_EEPROM(253+i);
      }
    }
    
    for(i=0; i< 3; i++) {    
        colors[i] = cols[i];
        real_colors[i] = cols[i];
    }
    
    
    

    INTCONbits.GIE = 1;
    INTCONbits.PEIE = 1;
    RCONbits.IPEN = 1;

    IPR1bits.TMR2IP = 0;
    
    
    
    // init timers
    OpenTimer0(T0_16BIT & T0_SOURCE_INT & T0_PS_1_1);
    OpenTimer1(T1_16BIT_RW & T1_PS_1_1 & T1_OSC1EN_OFF & T1_SOURCE_INT);
    ot2(TIMER_INT_ON & T2_PS_1_16 & T2_POST_1_16);
    OpenTimer3(T3_16BIT_RW & T3_PS_1_1 & T3_OSC1EN_OFF & T3_SOURCE_INT);
}

void ProcessIO(void)
{   
    if ((usb_device_state < CONFIGURED_STATE)||(UCONbits.SUSPND==1)) {
        osc = 0;
        return;    
    }
    ServiceRequests();  
}


void ServiceRequests(void)
{
    int i;
  
    oscilloscope();    
    
    counter = 0;
    if(USBGenRead((byte*)&dataPacket,sizeof(dataPacket)))
    {
        switch(dataPacket.cmd)
        {
            case 0:
                 dataPacket._byte[0] = MAJOR_VERSION;
                 dataPacket._byte[1] = colors[0];
                 dataPacket._byte[2] = colors[1];
                 dataPacket._byte[3] = colors[2];
                 counter=0x04;
                 break;
            case 37:
            
                for(i=0; i< 3; i++) {
                    dataPacket._byte[i*4+0] = real_colors[i];
                    dataPacket._byte[i*4+1] = cols[state*3+i];
                    dataPacket._byte[i*4+2] = colors[i];
                    dataPacket._byte[i*4+3] = 111;
                
            
                 }
                 counter=12;
                 break;
            case 1:
                 endpoint_wait_counter = 1;
                 memcpy(real_colors, dataPacket.colors, 3);
                 dataPacket._byte[1] = 233; //OK
                 counter = 4;
                 break;
            case 14:
                osc_sleeptime = dataPacket._byte[1];
                osc_sleeptime = osc_sleeptime * 256  + dataPacket._byte[2];
                break;
            case 15:
                ADCON0bits.ADON = 1;
                osc = 1;
                CloseTimer0();
                CloseTimer1(); // !!!
                IPR1bits.TMR1IP = 0;
                CloseTimer2();
                CloseTimer3();
                //ot2(TIMER_INT_ON & T2_PS_1_1 & T2_POST_1_16);
                rpin = 0;
                gpin = 0;
                bpin = 0;
                osc_fill=0;
                osc_sent=(OSC_ARRAYS-1);
                osc_pos=0;
                osc_overflow = 0;
                osc_seq = 0;
                osc_packets = 0;

                
                //enable interrupt
                PIE1bits.ADIE = 1;
                IPR1bits.ADIP = 0; //low-priority interrupt
                
                // a/d syklit = 1792 per sample
                // todelliset syklit per sample = 3097

                ADCON1 = 0b00001110;
                ADCON2 = 0b10110110;
                ADCON0bits.GO = 1;
                
                dataPacket._byte[0] = MAJOR_VERSION;
                dataPacket._byte[1] = 233;
                dataPacket._byte[2] = osc_sleeptime >> 8;
                dataPacket._byte[3] = osc_sleeptime & 0xff;
                counter = 4;
                break;
            case 16:
                osc = 0;
                current_osc[OSC_BYTES-1] = 255;
                if(!mUSBGenTxIsBusy())
                    USBGenWrite((byte*)&current_osc, OSC_BYTES);
                break;
            case 6:
                // change settings accordingly
                speed = dataPacket.speed;
                endpoint_wait_time = dataPacket.wait;
                colam = dataPacket.colAmount;
                if (colam > MAX_COL_AM)
                  colam = MAX_COL_AM;
                memcpy(cols, dataPacket.colArray, dataPacket.colAmount*3); 
                dataPacket._byte[0] = MAJOR_VERSION;
                dataPacket._byte[1] = 233; //OK
                counter = 4;
                break;
            case 7:
                // send back current color settings.
                dataPacket.speed = speed;
                dataPacket.wait = endpoint_wait_time;
                if (colam > MAX_COL_AM)
                  colam = MAX_COL_AM;
                dataPacket.colAmount = colam;
                //memcpy(dataPacket.colArray, cols, dataPacket.colAmount*3); 
                for(i=0; i< colam*3; i++) {
                    dataPacket.colArray[i] = cols[i];
                }
                dataPacket._byte[1] = 233; //OK
                counter = sizeof(DATA_PACKET);
                break;
            case 8:
                // transfer data packet contents to the EEPROM
                // change settings accordingly
                dataPacket._byte[1] = 233; //OK
                if(!mUSBGenTxIsBusy())
                    USBGenWrite((byte*)&dataPacket,4);
                updateEEPROM();
                break;

              case 10:
                // save white balance
                for(i=0; i< 3; i++) {
                  write_EEPROM(253+i, dataPacket._byte[1+i]);
                }
                write_EEPROM(252, 178);
              case 11:
                // set white balance
                for(i=0; i< 3; i++) {
                    white_balance[i] = dataPacket._byte[1+i];
                }
              case 12:
                // get white balance
                for(i=0; i< 3; i++) {
                  dataPacket._byte[2+i] = white_balance[i];
                }
                dataPacket._byte[1] = 233; //OK
                counter = 5;
                break;
            default:
                 break;
        }
        
        
        
        if(counter != 0)
        {
            if(!mUSBGenTxIsBusy())
               USBGenWrite((byte*)&dataPacket,counter);
        }
    }

}
